-- Initial file of "Toolbox" mod.
-- Version 4.2
-- Feel free to study and use any part of the source code.
--
-- Made by ImpassIve_rus


--
-- PUBLIC FUNCTIONS

-- RegisterMod ( name, version )
-- RegisterMod ( name, version, folder )
-- IsModRegistered ( name ) => true/false
-- GetRegisteredMod ( name ) => require(name) / nil
-- GetModVersion ( name ) => version / nil
-- IsModEnabled ( name ) => true / false / (nil if the mod was not found)

-- AL_AddListener ( action_id, function_id, function )
-- AL_DeleteListener ( action_id, function_id )
-- AL_SendAction ( id, parameter )

-- Default Actions:
-- GAME_PREPARE - called before dungeon generation
-- GAME_START - called when the match starts
-- RESPAWN - called when knight is preparing to spawn. Gives a player as parameter of the function.
-- MOD_REGISTERED - called every time some mod was registered. Gives a name of the mod as a first argument.


-- RegisterHook( hook , id, function ) => true (success) / false (fail)
-- DeleteHook( hook , id )
---- Correct hook types are: ("WEAPON_DOWNSWING"/"WEAPON_PARRY"/"KNIGHT_DAMAGE"/"CREATURE_SQUELCH"/"SHOOT"/"MISSILE_MISS")

module(...)

dofile("config.lua")

-- LOCAL AND PUBLIC VARIABLES
local C = require("classic")
local mod_list = {{"toolbox", 4, ..., state="on"}} --ID, version, module name, state (is_enabled)

local action_running = false
local a_listeners = {} --listeners={id, function}
local listeners_toDelete = {}

local h_W_D = {{"stdhook", kts.HOOK_WEAPON_DOWNSWING, true}} --ID, function, is_enabled
local h_W_P = {{"stdhook", kts.HOOK_WEAPON_PARRY, true}}
local h_K_D = {{"stdhook", kts.HOOK_KNIGHT_DAMAGE, true}}
local h_C_S = {{"stdhook", kts.HOOK_CREATURE_SQUELCH, true}}
local h_S = {{"stdhook", kts.HOOK_SHOOT, true}}
local h_M_M = {{"stdhook", kts.HOOK_MISSILE_MISS, true}}
local old_start_game = kts.MENU.start_game_func
local old_prepare_game = kts.MENU.prepare_game_func


local hook_running = false
local hook_toDelete = {}

local current_menu_mod = nil
local last_menu_state = "on"
local menumods = {}


-- LOCAL FUNCTIONS
local function DeleteHooks()
	for _i=1,#hook_toDelete do
		DeleteHook(hook_toDelete[_i][1], hook_toDelete[_i][2])
	end	
	hook_toDelete = {}
end

local function DeleteListeners()
	for _i=1,#listeners_toDelete do
		AL_DeleteListener(listeners_toDelete[_i][1], listeners_toDelete[_i][2])
	end	
	listeners_toDelete = {}
end

local function invoke_hooks(hooks_list)
	hook_running = true
	for _i=1,#hooks_list do
		if hooks_list[_i][3] then
			hooks_list[_i][2]()
		end
	end
	hook_running = false
	DeleteHooks()
end

local function hook_WEAPON_DOWNSWING() 
	invoke_hooks(h_W_D)
end

local function hook_WEAPON_PARRY() 
	invoke_hooks(h_W_P)
end

local function hook_KNIGHT_DAMAGE() 
	invoke_hooks(h_K_D)
end

local function hook_CREATURE_SQUELCH() 
	invoke_hooks(h_C_S)
end

local function hook_SHOOT() 
	invoke_hooks(h_S)
end

local function hook_MISSILE_MISS() 
	invoke_hooks(h_M_M)
end

local function start_game()
	if old_start_game then
		old_start_game()
	end
	AL_SendAction("GAME_START", nil)
end

local function resp(player, fier)
	return AL_SendAction("RESPAWN", player);
end

local function prepare_game()
	if old_prepare_game then
		old_prepare_game()
	end
	kts.SetRespawnFunction(resp)
	AL_SendAction("GAME_PREPARE", nil)
end

local function GetModTable(name)
	name = string.lower(name)
	for _i=1,#mod_list do
		if mod_list[_i][1] == name then
			return mod_list[_i]
		end
	end
	return nil
end

local function SendStateFuncs(items, S)
	for _i=1,#items do
		local item = GetModTable(items[_i])
		if item.state == "on" then
			item[3].toolbox_enable(S)
		else
			item[3].toolbox_disable(S)
		end	
	end
end

-- PUBLIC FUNCTIONS

-- RegisterMod ( name, version )
-- RegisterMod ( name, version, folder )
-- IsModRegistered ( name ) => true/false
-- GetRegisteredMod ( name ) => require(name) / nil
-- GetModVersion ( name ) => version / nil
-- IsModEnabled ( name ) => true / false /(nil if the mod was not found)

-- AL_AddListener ( action_id, function_id, function )
-- AL_DeleteListener ( action_id, function_id )
-- AL_SendAction ( id, parameter )

-- Default Actions:
-- GAME_PREPARE - called before dungeon generation
-- GAME_START - called when the match starts
-- RESPAWN - called when knight is preparing to spawn. Gives a player as parameter of the function.
-- MOD_REGISTERED - called every time some mod was registered. Gives a name of the mod as a first argument.


-- RegisterHook( hook , id, function ) => true (success) / false (fail)
-- DeleteHook( hook , id )
---- Correct hook types are: ("WEAPON_DOWNSWING"/"WEAPON_PARRY"/"KNIGHT_DAMAGE"/"CREATURE_SQUELCH"/"SHOOT"/"MISSILE_MISS")


function AL_AddListener ( action_id, function_id, func )
	local h_table = a_listeners[action_id]
	if (h_table==nil) then
		h_table = { {function_id, func} }
		a_listeners[action_id] = h_table
		return
	end
	
	for _j=1,#h_table do
		if (h_table[_j][1] == function_id) then
			h_table[_j][2] = func
			return
		end
	end	
	table.insert(h_table, {function_id, func})
end


function AL_SendAction ( id, parameter )
	local h_table = a_listeners[id]
	if (h_table==nil) then
		a_listeners[id] = {}
		return nil
	end
	action_running = true
	
	local result = nil
	for _j=1,#h_table do
		local res = h_table[_j][2](parameter)
		if res == "retry" then
			if (result == nil) then
				result = "retry"
			end
		elseif res ~= nil then
			result = res
		end
	end	
	action_running = false
	DeleteListeners()
	return result
end

function AL_DeleteListener ( action_id, function_id )
	local h_table = a_listeners[action_id]
	if (h_table==nil) then
		return
	end
	if action_running then
		table.insert(listeners_toDelete, {action_id, function_id})
	else	
		for _j=1,#h_table do
			if (h_table[_j][1] == function_id) then
				table.remove(h_table, _j)
				break
			end
		end	
	end
end

function RegisterMod(name, version, folder)
	if (folder == nil) then
		folder = name
	end	
	if (package.loaded[folder]) then
		table.insert(mod_list, {string.lower(name), version, require(folder), state=DEFAULT_MOD_STATE})
	else
		table.insert(mod_list, {string.lower(name), version, nil, state=DEFAULT_MOD_STATE})
	end
	
	AL_SendAction("MOD_REGISTERED", name)
end

function RegisterNamedMod(name, version, folder)
	RegisterMod(name, version, folder)
end

function IsModRegistered(name)
	name = string.lower(name)
	for _i=1,#mod_list do
		if mod_list[_i][1] == name then
			return true
		end
	end
	return false
end

function GetRegisteredMod(name)
	name = string.lower(name)
	for _i=1,#mod_list do
		if mod_list[_i][1] == name then
			return mod_list[_i][3]
		end
	end
	return nil
end

function IsModEnabled (name) 
	name = string.lower(name)
	for _i=1,#mod_list do
		if mod_list[_i][1] == name then
			return mod_list[_i][4]
		end
	end
	return nil
end

function GetModVersion(name)
	name = string.lower(name)
	for _i=1,#mod_list do
		if mod_list[_i][1] == name then
			return mod_list[_i][2]
		end
	end
	return nil
end

function RegisterHook(hook, id, funct, is_parallel)
	local h_table
	if (hook == "WEAPON_DOWNSWING") then
		h_table = h_W_D
	elseif (hook == "WEAPON_PARRY") then
		h_table = h_W_P
	elseif (hook == "KNIGHT_DAMAGE") then
		h_table = h_K_D
	elseif (hook == "CREATURE_SQUELCH") then
		h_table = h_C_S
	elseif (hook == "SHOOT") then
		h_table = h_S
	elseif (hook == "MISSILE_MISS") then
		h_table = h_M_M
	else
		return false
	end
	for _i=1,#h_table do
		if (h_table[_i][1]==id) then
			h_table[_i][2] = funct
			h_table[_i][3] = true
			return true
		end	
	end
	table.insert(h_table, {id, funct, true})
	return true
end

function DeleteHook(hook, id)
	local h_table
	if (hook == "WEAPON_DOWNSWING") then
		h_table = h_W_D
	elseif (hook == "WEAPON_PARRY") then
		h_table = h_W_P
	elseif (hook == "KNIGHT_DAMAGE") then
		h_table = h_K_D
	elseif (hook == "CREATURE_SQUELCH") then
		h_table = h_C_S
	elseif (hook == "SHOOT") then
		h_table = h_S
	elseif (hook == "MISSILE_MISS") then
		h_table = h_M_M
	else
		return
	end
	
	for _i=1,#h_table do
		if (h_table[_i][1]==id) then
			if hook_running then
				h_table[_i][3] = false --Disable hook now, will delete it later
				table.insert(hook_toDelete, {hook, id})
			else	
				table.remove(h_table, _i)
			end
			break
		end	
	end
end

--ON INIT

kts.HOOK_WEAPON_DOWNSWING = hook_WEAPON_DOWNSWING
kts.HOOK_WEAPON_PARRY = hook_WEAPON_PARRY
kts.HOOK_KNIGHT_DAMAGE = hook_KNIGHT_DAMAGE
kts.HOOK_CREATURE_SQUELCH = hook_CREATURE_SQUELCH
kts.HOOK_SHOOT = hook_SHOOT
kts.HOOK_MISSILE_MISS = hook_MISSILE_MISS
kts.MENU.start_game_func = start_game
kts.MENU.prepare_game_func = prepare_game

local function add_entry(name)
	local mmod = GetRegisteredMod(name)
	if (mmod == nil) then
		return
	end	
	if (mmod.toolbox_enable == nil) or (mmod.toolbox_disable == nil) then
		return
	end
	table.insert(menumods, name)
	if not MENU_ENABLED then
		--if (menu_enabled == nil) then
		--	old_onselect = kts.MENU.on_select
		--	kts.MENU.on_select = onselect
		--	menu_enabled = true
		--end
		return
	end	
	local mod_table = GetModTable(name)
	current_menu_mod = mod_table
	local Menu = require("menu")
	if (menu_enabled==nil) then
		local toolmenu1 = {
			id = "toolbox_1",
			text = "Mod",
			choices = {}
		}
		local toolmenu2 = {
			id = "toolbox_2",
			text = "State",
			choices = {
				{
					id = "on",
					text = "Enable",
					constrain = function (S)
						if (current_menu_mod[1] ~= S.toolbox_1) then
							current_menu_mod = GetModTable(S.toolbox_1)
							S.toolbox_2 = current_menu_mod.state
						else	
							current_menu_mod.state = "on"
							last_menu_state = "on"
						end
						SendStateFuncs(menumods, S)	
					end
				},
				{
					id = "off",
					text = "Disable",
					constrain = function (S)
						if (current_menu_mod[1] ~= S.toolbox_1) then
							current_menu_mod = GetModTable(S.toolbox_1)
							S.toolbox_2 = current_menu_mod.state
						else
							current_menu_mod.state = "off"
							last_menu_state = "off"
						end
						SendStateFuncs(menumods, S)	
					end
				}
			}
		}
		menu_enabled = true
		Menu.add_menu_item("spacer")
		Menu.add_menu_item(toolmenu1)
		Menu.add_menu_item(toolmenu2)
	end
	
	local miss = {
		id = name,
		text = name,
		constrain = function (S)
			current_menu_mod = GetModTable(S.toolbox_1)
			if (last_menu_state == S.toolbox_2) then
				S.toolbox_2 = current_menu_mod.state
			end
		end
	}
	Menu.add_menu_choice("toolbox_1", miss)
end

AL_AddListener("MOD_REGISTERED", "Toolbox_mr", add_entry)